package pt.uc.dei.rest_api_robustness_tester;

import net.openhft.compiler.CompilerUtils;
import org.apache.http.HttpHost;
import pt.uc.dei.rest_api_robustness_tester.faultload.Faultload;
import pt.uc.dei.rest_api_robustness_tester.faultload.FaultloadExecutor;
import pt.uc.dei.rest_api_robustness_tester.faultload.FaultloadExecutorConfig;
import pt.uc.dei.rest_api_robustness_tester.faultload.FaultloadGenerator;
import pt.uc.dei.rest_api_robustness_tester.specification.RestApi;
import pt.uc.dei.rest_api_robustness_tester.specification.RestApiSetup;
import pt.uc.dei.rest_api_robustness_tester.utils.Config;
import pt.uc.dei.rest_api_robustness_tester.utils.XlsxWriter;
import pt.uc.dei.rest_api_robustness_tester.workload.*;

import java.io.File;
import java.io.PrintStream;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.LinkedHashMap;
import java.util.Map;

public class RestApiRobustnessTester
{
    private void Init(String[] args)
    {
        try
        {
            Map<String, String> argsMap = new LinkedHashMap<>();
            String lastArg = null;
            for(String arg : args)
            {
                if(arg.startsWith("--"))
                {
                    lastArg = arg;
                    argsMap.put(lastArg, "");
                }
                else
                {
                    if(lastArg != null)
                        argsMap.put(lastArg, arg);
                }
            }
            new Config(argsMap);
        }
        catch(Exception e)
        {
            System.out.println("Invalid arguments!");
            System.out.println("Argument syntax: [options...] --api-file <API *.java file> --wl-results <workload results *.xlsx file>" +
                    " --fl-results <faultload results *.xlsx file>");
            System.out.println("Available options:");
            System.out.println("\t--wl-rep <reps> Number of parameter repetitions for workload generation (default is 10)");
            System.out.println("\t--wl-retry <retries> Number of retries for failed workload requests (default is 1)");
            System.out.println("\t--wl-max-time <seconds> Maximum duration of workload execution or 0 for exhausting all requests (default is 0)");
            System.out.println("\t--fl-rep <reps> Maximum number of times a fault can be injected into a request parameter (default is 3)");
            System.out.println("\t--fl-max-time <seconds> Maximum duration of faultload execution or 0 for exhausting all requests (default is 0)");
            System.out.println("\t--out <file name> Print output to an external file rather than to the console");
            System.out.println("\t--keep-failed-wl Keeps the workload requests that result in a non 2xx response");
            System.out.println("\t--proxy-port <port> Run as a fault injector proxy (i.e., only faultload execution) bound to this port");
            System.out.println("\t--conn-timeout <seconds> Timeout value for all connections (default is 60)");
            System.exit(0);
        }
    
        if(Config.Instance().out != null)
        {
            try
            {
                PrintStream output = new PrintStream(new File(Config.Instance().out));
                System.setOut(output);
                System.setErr(output);
            }
            catch(Exception e)
            {
                System.err.println("Could not redirect output to file " + Config.Instance().out);
            }
        }
    }
    
    private void Run() throws Exception
    {
        if(!Config.IsInitialized())
            throw new Exception("A Config must be defined before running");
        
        String apiSetupSourceCode = new String(Files.readAllBytes(Paths.get(Config.Instance().apiClassFile)));
        String apiSetupClassName = new File(Config.Instance().apiClassFile.replace(".java", "")).getName();
        Class cl = CompilerUtils.CACHED_COMPILER.loadFromJava(apiSetupClassName, apiSetupSourceCode);
        RestApiSetup restApiSetup = (RestApiSetup) cl.newInstance();
        RestApi restAPI = restApiSetup.Load();
        
        restAPI.Specification().Validate();
        
        System.out.println(restAPI.Specification().apiNameVersion);
    
        if(!restAPI.Specification().securitySchemes.isEmpty())
        {
            System.out.println("Security schemes:");
            for (String s : restAPI.Specification().securitySchemes.keySet())
                System.out.println("\t" + s + ": " + restAPI.Specification().securitySchemes.get(s).type);
        }
        else
            System.out.println("No security schemes defined");
    
        if(Config.Instance().proxyPort == Config.UNKNOWN_PORT)
        {
            WorkloadGeneratorConfig generatorConfig = new WorkloadGeneratorConfig();
            generatorConfig.RepetitionsPerParameter(Config.Instance().WL_REP);
    
            WorkloadGenerator workloadGenerator = new WorkloadGenerator(restAPI, generatorConfig);
            Workload workload = workloadGenerator.Generate();
    
            System.out.println("Workload requests generated: " + workload.WorkloadRequests().size());
    
            XlsxWriter workloadWriter = new XlsxWriter(new File(Config.Instance().workloadResultsFile));
    
            WorkloadExecutorConfig workloadConfig = new WorkloadExecutorConfig();
            workloadConfig.MaxRetriesOnFail(Config.Instance().WL_RETRY).
                    ExecuteRandomly().
                    WriteTo(workloadWriter);
            if(Config.Instance().WL_MAX_TIME <= 0)
                workloadConfig.StopWhenAllDone();
            else
                workloadConfig.StopWhenTimeEnds(Config.Instance().WL_MAX_TIME * 1000);
            if (Config.Instance().keepFailedWorkload)
                workloadConfig.KeepFailedRequests();
            else
                workloadConfig.DiscardFailedRequests();
    
            System.out.println("***************** Workload started *****************");
            if(Config.Instance().WL_MAX_TIME <= 0)
                System.out.println("Stop when workload is exhausted");
            else
                System.out.println("Stop when " + Config.Instance().WL_MAX_TIME + " seconds pass");
    
            LocalWorkloadExecutor workloadExecutor = new LocalWorkloadExecutor(workload, restAPI, workloadConfig);
            //FIXME: temporarily skipping workload
            //workloadExecutor.Execute();
            Workload validWorkload = workload;  //workloadExecutor.GetResults();
            
            workloadWriter.Write(restAPI.Name());
    
            System.out.println("***************** Workload finished *****************");
    
            FaultloadGenerator faultloadGenerator = new FaultloadGenerator();
            Faultload faultload = faultloadGenerator.Generate();
    
            XlsxWriter faultloadWriter = new XlsxWriter(new File(Config.Instance().faultloadResultsFile));
    
            FaultloadExecutorConfig faultloadConfig = new FaultloadExecutorConfig();
            faultloadConfig.MaxInjectionsPerFault(Config.Instance().FL_REP).
                    WriteTo(faultloadWriter);
    
            FaultloadExecutor faultloadExecutor = new FaultloadExecutor(faultload, restAPI, faultloadConfig);
    
            workloadConfig = new WorkloadExecutorConfig();
            workloadConfig.ExecuteRandomly().
                    Hook(faultloadExecutor).
                    WriteTo(faultloadWriter);
            if(Config.Instance().FL_MAX_TIME <= 0)
                workloadConfig.StopWhenAllDone();
            else
                workloadConfig.StopWhenTimeEnds(Config.Instance().FL_MAX_TIME * 1000);
    
            System.out.println("Workload requests to use for fault injection: " + validWorkload.WorkloadRequests().size());
            System.out.println("Total injectable faults: " + faultloadExecutor.FaultsLeft());
            Map<String, Integer> faultsPerOperation = faultloadExecutor.FaultsLeftPerOperation();
            for(String op : faultsPerOperation.keySet())
                System.out.println("\t" + op + ": " + faultsPerOperation.get(op) + " faults");
            
            System.out.println("***************** Workload + Faultload started *****************");
            if(Config.Instance().FL_MAX_TIME <= 0)
                System.out.println("Stop when all faults are exhausted");
            else
                System.out.println("Stop when " + Config.Instance().FL_MAX_TIME + " seconds pass");
    
            workloadExecutor = new LocalWorkloadExecutor(validWorkload, restAPI, workloadConfig);
            workloadExecutor.Execute();
            
            faultloadWriter.Write(restAPI.Name());
    
            System.out.println("***************** Workload + Faultload finished *****************");
        }
        else
        {
            System.out.println("Starting proxy fault injector at http://localhost:" + Config.Instance().proxyPort);
    
            HttpHost proxyHost = new HttpHost("localhost", Config.Instance().proxyPort, "http");
    
            FaultloadGenerator faultloadGenerator = new FaultloadGenerator();
            Faultload faultload = faultloadGenerator.Generate();
    
            XlsxWriter faultloadWriter = new XlsxWriter(new File(Config.Instance().faultloadResultsFile));
    
            FaultloadExecutorConfig faultloadConfig = new FaultloadExecutorConfig();
            faultloadConfig.MaxInjectionsPerFault(Config.Instance().FL_REP).
                    WriteTo(faultloadWriter);
    
            FaultloadExecutor faultloadExecutor = new FaultloadExecutor(faultload, restAPI, faultloadConfig);
    
            WorkloadExecutorConfig workloadConfig = new WorkloadExecutorConfig();
            workloadConfig.ExecuteRandomly().
                    ProxyHost(proxyHost).
                    Hook(faultloadExecutor).
                    WriteTo(faultloadWriter);
            if(Config.Instance().FL_MAX_TIME <= 0)
                workloadConfig.StopWhenAllDone();
            else
                workloadConfig.StopWhenTimeEnds(Config.Instance().FL_MAX_TIME * 1000);
    
            System.out.println("Total injectable faults: " + faultloadExecutor.FaultsLeft());
            Map<String, Integer> faultsPerOperation = faultloadExecutor.FaultsLeftPerOperation();
            for(String op : faultsPerOperation.keySet())
                System.out.println("\t" + op + ": " + faultsPerOperation.get(op) + " faults");
    
            System.out.println("***************** Proxy faultload started *****************");
            if(Config.Instance().FL_MAX_TIME <= 0)
                System.out.println("Stop when all faults are exhausted");
            else
                System.out.println("Stop when " + Config.Instance().FL_MAX_TIME + " seconds pass");
    
            RemoteWorkloadExecutor workloadExecutor = new RemoteWorkloadExecutor(restAPI, workloadConfig);
            workloadExecutor.Execute();
    
            faultloadWriter.Write(restAPI.Name());
    
            System.out.println("***************** Proxy faultload finished *****************");
        }
    }
    
    public static void main(String[] args)
    {
        try
        {
            RestApiRobustnessTester tool = new RestApiRobustnessTester();
            tool.Init(args);
            tool.Run();
        }
        catch(Exception e )
        {
            System.err.println("Error: " + e);
            e.printStackTrace();
        }
    }
}
